#' Function for learning Hybrid Probabilistic Decision Graph (HPDG) classifiers from data
#'
#' Learns an Hybrid Probabilistic Decision Graph (HPDG) classifier from a given dataset
#'
#' @param dataset a String indicating the data file (the path has to be included also). Data columns
#'        are separated by ";" and a header indicates the variables names.
#'
#' @param method the method applied to build the classifier:
#'    \itemize{
#'      \item{ranked}{The variables are ranked by mutual information and they are connected as:
#'                    C -> X_1 -> X_2 -> X_3 -> ...}
#'      \item{rankedCR}{The variables are ranked by mutual information and then each X_i is
#'                    connected to the leaf that maximizes the CR of the classifier wrt to a
#'                    test database.}
#'      \item{CR}{In each step a non included variable Xj is connected to a leaf variable Xi
#'                    satisfying that <X_i, X_j> = arg max CR(X_i,X_j,pdg,train,test). This is
#'                    computationally the hardest option. See Nielsen et. al (2009) - Algorithm 2.}
#'      \item{NB}{Variables are arranged following a NB structure.}
#'      \item{ChowLiu}{Variables are arranged in a TAN structure given by the Chow Liu algorithm.}
#'    }
#'
#' @param densityType the type of density ("MOP" or "MTE").
#'
#' @param maxIntervals maximum number of intervals in which the continuous parents are split.
#'
#' @param collapseRate A number k in the formula nCases = k * R(Xi), where R(Xi) is the number of
#'                states in the variable Xi and nCases is the minimum number of cases to be
#'                allowed to learn the parameters without collapsing nodes. By default this
#'                number is set to 0.5.
#'
#' @param mergeRate A number between 0 and 1 indicating the percentage of parameters to be checked
#'             for merging (non mandatory argument). By default this argument is set to 0.5.
#'
#' @param maxTermsDensity the maximum number of terms to be fit in the density.
#'
#' @return The HPDG classifier learnt from data.
#'
#' @seealso \code{chow.liu} in package \code{bnlearn}
#' @export
#' @examples
#' \dontrun{
#' ## Learns a hpdg classifier from a dataset
#' classifier <- hpdgclassifier("data/iris.csv", "rankedCR", "MOP", 3, 2, 0.5, 5)
#' }
#'
hpdgclassifier <- function (dataset, method=c("ranked", "rankedCR", "CR", "NB", "ChowLiu"), densityType=c("MOP", "MTE"), maxIntervals, collapseRate, mergeRate, maxTermsDensity) {

  # Checks arguments
  if(is.null(dataset)) stop ("Please specify the path including the dataset.")
  method <- match.arg(method)
  densityType <- match.arg(densityType)
  if(mergeRate<0 || mergeRate>1) stop("\nMerge rate must be between 0 and 1.")

  # Loads the dataset
  db = Database$new(dataset)

  # Initializes the PDG
  pdg = PDGClassifier$new(
    database = db,
    maxIntervals=maxIntervals,
    collapseRate = collapseRate,
    mergeRate=mergeRate,
    method=method,
    densityType=densityType,
    maxTermsDensity=maxTermsDensity)

  # Learns the classifier
  pdg$learn()

  #pdg$show()
  return(pdg)
}


#' Function for applying an Hybrid Probabilistic Decision Graph classifier (HPDG) to each
#' observation in a dataset. The HPDG classifier predicts a class for each observation.
#'
#' @param hpdg_classifier an Hybrid Probabilistic Decision Graph classifier.
#'
#' @param dataset a String indicating the data file (the path has to be included also).
#'        Data columns are separated by ";" and a header indicates the variables names.
#'
#' @return Return a vector of predicted values for the class variable, one for each observation in
#          the given dataset
#'
#' @seealso \code{\link{hpdgclassifier}}, \link{score}
#' @export
#' @examples
#' \dontrun{
#' ## Learns a hpdg classifier from a dataset
#' classifier <- hpdgclassifier("data/iris.csv", "rankedCR", "MOP", 3, 2, 0.5, 5)
#' predicted_values <- classify (classifier, "data/iris.csv")
#' }
#'
classify <- function(hpdg_classifier, dataset) {
  # Checks arguments
  if (class(hpdg_classifier) != "PDGClassifier") stop ("'hpdg classifier' is missing or incorrect")
  if(is.null(dataset)) stop ("Please specify the path including the dataset.")

  # Loads the dataset
  db = Database$new(dataset)

  # Makes predictions
  predictedValues <- hpdg_classifier$predict(db)

  return (predictedValues)
}

#' Function for applying an Hybrid Probabilistic Decision Graph classifier (HPDG) to each
#' observation in a dataset. The HPDG classifier predicts a class for each observation and
#' return its probability value.
#'
#' @param hpdg_classifier an Hybrid Probabilistic Decision Graph classifier.
#'
#' @param dataset a String indicating the data file (the path has to be included also).
#'        Data columns are separated by ";" and a header indicates the variables names.
#'
#' @return Return a vector with the probability values associated to the predicted values
#' for the class variable, one for each observation in the given dataset
#'
#' @seealso \code{\link{hpdgclassifier}}, \link{classify}
#' @export
#' @examples
#' \dontrun{
#' ## Learns a hpdg classifier from a dataset
#' classifier <- hpdgclassifier("data/iris.csv", "rankedCR", "MOP", 3, 2, 0.5, 5)
#' predicted_values <- classify (classifier, "data/iris.csv")
#' predicted_scores <- score(classifier, "data/iris.csv")
#' }
#'
score <- function(hpdg_classifier, dataset) {
  # Checks arguments
  if (class(hpdg_classifier) != "PDGClassifier") stop ("'hpdg classifier' is missing or incorrect")
  if(is.null(dataset)) stop ("Please specify the path including the dataset.")

  # Loads the dataset
  db = Database$new(dataset)

  # Makes predictions
  predictedScores <- hpdg_classifier$score(db)

  return (predictedScores)
}

#' Function for applying an Hybrid Probabilistic Decision Graph classifier (HPDG) to each
#' observation in a dataset. The HPDG classifier computes the class distribution for each
#' observation and returns it.
#'
#' @param hpdg_classifier an Hybrid Probabilistic Decision Graph classifier.
#'
#' @param dataset a String indicating the data file (the path has to be included also).
#'        Data columns are separated by ";" and a header indicates the variables names.
#'
#' @return Return a matrix with the class probability distribution associated to
#' each observation in the given dataset
#'
#' @seealso \code{\link{hpdgclassifier}}, \link{classify}, \link{score}
#' @export
#' @examples
#' \dontrun{
#' ## Learns a hpdg classifier from a dataset
#' classifier <- hpdgclassifier("data/iris.csv", "rankedCR", "MOP", 3, 2, 0.5, 5)
#' predicted_values <- classify (classifier, "data/iris.csv")
#' predicted_scores <- score (classifier, "data/iris.csv")
#' predicted_distributions <- scores(classifier, "data/iris.csv")
#' }
#'
scores <- function(hpdg_classifier, dataset) {
  # Checks arguments
  if (class(hpdg_classifier) != "PDGClassifier") stop ("'hpdg classifier' is missing or incorrect")
  if(is.null(dataset)) stop ("Please specify the path including the dataset.")

  # Loads the dataset
  db = Database$new(dataset)

  # Makes predictions
  predictedClassDistribution <- hpdg_classifier$scores(db)

  return (predictedClassDistribution)
}


#' Function for evaluating an Hybrid Probabilistic Prediction Graph (HPDG) on a dataset.
#' The HPDG classifier was previously builded and is applied to predict a class value
#' for each observation in the given dataset. The evaluation is performed comparing the
#' each predicted class value with the real class value given in the dataset.
#'
#' The function displays the evaluation measures computed using these real an predicted
#' class values.
#'
#' @param hpdg_classifier an Hybrid Probabilistic Decision Graph classifier.
#'
#' @param dataset a String indicating the data file (the path has to be included also).
#'        Data columns are separated by ";" and a header indicates the variables names.
#'
#' @return A classification matrix.
#'
#' @seealso \code{\link{hpdgclassifier}}, \code{\link{classificationMatrix}}
#' @export
#' @examples
#' \dontrun{
#' ## Learns a hpdg classifier from a dataset
#' classifier <- hpdgclassifier("data/iris.csv", "rankedCR", "MOP", 3, 2, 0.5, 5)
#' evaluate (classifier, "data/iris.csv")
#' }
#'
evaluate <- function (hpdg_classifier, dataset) {
  # Checks arguments
  if (class(hpdg_classifier) != "PDGClassifier") stop ("'hpdg classifier' is missing or incorrect")
  if(is.null(dataset)) stop ("Please specify the path including the dataset.")

  # Loads dataset and make predictions
  predictedValues <- classify(hpdg_classifier, dataset)

  # Gets the real class values from the  loaded dataset
  db <- hpdg_classifier$getDatabase()
  realValues <- db$getColumnData(hpdg_classifier$getRoot()) # db$getColumnDataAt(classifier$getClassIndex())

  # Evaluates the classifier
  clabels <- db$getStates(db$getNumberOfVars())
  cm <- classificationMatrix(clabels, realValues, predictedValues)

  # Displays results
  cat("\n\n*****************************************************")
  cat("\n\nHPDGC Evaluation\n")
  cat("\nHPDGC model size =", hpdg_classifier$size(),"\n")
  summarize(cm)

  return(cm)
}

#' Function for performing k-fold cross validation of an Hybrid Probabilistic Decision
#' Graph (HPDG) classifier on a dataset. The HPDG Classifier is learnt on the whole
#' data set. The evaluation results displayed are obtained doing the k-fold cross
#' validation process.
#'
#' @param folds the number of folds to be used. It has to be greater than 1. By default is set to 5.
#'
#' @param dataset a String indicating the data file (the path has to be included also). Data columns
#'        are separated by ";" and a header indicates the variables names.
#'
#' @param method the method applied to build the classifier:
#'   \itemize{
#'      \item{ranked}{The variables are ranked by mutual information and they are connected as:
#'                    C -> X_1 -> X_2 -> X_3 -> ...}
#'      \item{rankedCR}{The variables are ranked by mutual information and then each X_i is
#'                    connected to the leaf that maximizes the CR of the classifier wrt to a
#'                    test database.}
#'      \item{CR}{In each step a non included variable Xj is connected to a leaf variable Xi
#'                    satisfying that <X_i, X_j> = arg max CR(X_i,X_j,pdg,train,test). This is
#'                    computationally the hardest option. See Nielsen et. al (2009) - Algorithm 2.}
#'      \item{NB}{Variables are arranged following a NB structure.}
#'      \item{ChowLiu}{Variables are arranged in a TAN structure given by the Chow Liu algorithm.}
#'    }
#' @param densityType the type of density ("MOP" or "MTE").
#'
#' @param maxIntervals maximum number of intervals in which the continuous parents are split.
#'
#' @param collapseRate A number k in the formula nCases = k * R(Xi), where R(Xi) is the number of
#'                states in the variable Xi and nCases is the minimum number of cases to be
#'                allowed to learn the parameters without collapsing nodes. By default this
#'                number is set to 0.5.
#'
#' @param mergeRate A number between 0 and 1 indicating the percentage of parameters to be checked
#'             for merging (non mandatory argument). By default this argument is set to 0.5.
#'
#' @param maxTermsDensity the maximum number of terms to be fit in the density.
#'
#' @return Return an HPDG classifier learnt using the whole given dataset.
#'
#' @seealso \code{\link{hpdgclassifier}}, \code{\link{classificationMatrix}}
#' @export
#' @examples
#'  \dontrun{
#' ## Learns a hpdg classifier from a dataset and evaluates its performance applying
#' ## k-fold cross validation
#' classifier <- crossValidation(2, "data/iris.csv", "rankedCR", "MOP", 3, 2, 0.5, 5)
#' }
#'
crossValidation <- function (folds=NULL, dataset, method=c("ranked", "rankedCR", "CR", "NB", "ChowLiu"), densityType=c("MOP", "MTE"), maxIntervals, collapseRate, mergeRate, maxTermsDensity) {

  # Checks arguments
  if(is.null(folds)) folds = 5
  else
    if(folds<=1) {stop("\n The number of folds must be greater that 1.")}
  if(is.null(dataset)) stop ("Please specify the path including the dataset.")
  method <- match.arg(method)
  densityType <- match.arg(densityType)
  if(mergeRate<0 || mergeRate>1) stop("\nMerge rate must be between 0 and 1.")

  # Loads the dataset, gets the number of cases and the set of class labels
  db <- Database$new(dataset)
  nc <- db$getNumberOfCases()
  clabels <- db$getStates(db$getNumberOfVars())

  # Initializes results: confusion matrix, vector of times and vector of sizes
  ncl <- length(clabels)
  cm <- matrix (as.integer(0), nrow = ncl, ncol = ncl)
  rownames(cm) <- clabels # real class
  colnames(cm) <- clabels # predicted class
  timesv <- c()
  sizesv <- c()

  # Splits data in k folds, using k-1 for leraning and one for testing
  for (i in 1:folds){

    #train = db$getTrainCV(i,folds)
    #test = db$getTestCV(i,folds)

    train <- db$getTrainSet(i, folds)
    test <- db$getTestSet(i, folds)

    #cat("\n Train cases:",train$getNumberOfCases())
    #cat("\n Test cases:",test$getNumberOfCases())

    if ((train$getNumberOfCases()+test$getNumberOfCases())!= nc)
      stop("\nThe union of the train and test differs from the whole database")

    cat("\n*****************************************************")
    cat("\nBuilding HPDG classifier for partition",i,"\n")

    start.time <- Sys.time()
    # Initializes the PDG
    pdg = PDGClassifier$new(
      database = db,
      maxIntervals=maxIntervals,
      collapseRate = collapseRate,
      mergeRate=mergeRate,
      method=method,
      densityType=densityType,
      maxTermsDensity=maxTermsDensity)

    # Sets the train set
    pdg$setDatabase(train)

    # Learns the classifier
    pdg$learn()
    #pdg$show()
    #pdg$checkValidity()
    time <-  (Sys.time()- start.time)[[1]]

    # Evaluates the classifier on fold k
    testValues = test$getColumnData(pdg$getRoot()) # Gets test data
    cat("Evaluating HPDG classifier on fold",i,"\n")
    predictedValues = pdg$predict(test)            # Applies the classifier and makes predictions

    cmk <- classificationMatrix(clabels ,testValues, predictedValues)
    cm <- cm + cmk
    timesv <- c(timesv, time)
    sizesv <- c(sizesv, pdg$size())
  }

  # Builds classifier on whole dataset
  cat("\n*****************************************************")
  cat("\nBuilding HPDG classifier on whole dataset")
  start.time <- Sys.time()
  pdg <-  hpdgclassifier(dataset, method, densityType, maxIntervals, collapseRate, mergeRate, maxTermsDensity)
  time <- (Sys.time() - start.time)[[1]]
  cat("\nTime for building classifier =",time,"s.\n")

  # Displays results
  cat("\n\n*****************************************************")
  cat("\n------- ",folds,"-fold  Cross Validation  -------\n\n")
  cat("Average time to build the model =",mean(timesv),"(std.", stats::sd(timesv),")\n")
  cat("Average size of the model =",mean(sizesv),"(std.", stats::sd(sizesv),")\n")

  summarize(cm)
  return (pdg)
}

#' Function for showing the components of an Hybrid Probabilistic Decision Graph classifier (HPDG):
#' variables, structure and parameters.
#'
#' @param hpdg_classifier an Hybrid Probabilistic Decision Graph classifier.
#'
#' @return Return in console information about the estructure and parameters of an Hybrid
#' Probabilistic Decision Graph classifier (HPDG)
#'
#' @seealso \code{\link{hpdgclassifier}}
#' @export
#' @examples
#' \dontrun{
#' ## Exhibits the components of a hpdg classifier learned from a dataset
#' classifier <- hpdgclassifier("data/iris.csv", "rankedCR", "MOP", 3, 2, 0.5, 5)
#' exhibit(classifier)
#' }
#'
exhibit <- function(hpdg_classifier) {
  # Checks arguments
  if (class(hpdg_classifier) != "PDGClassifier") stop ("'hpdg classifier' is missing or incorrect")

  # Shows HPDG classifier components
  hpdg_classifier$show()
}


