##################################################################################################
######################################## CLASS DEFINITION ########################################
##################################################################################################
VariableNode <- setRefClass(
  Class = "VariableNode",
  fields = list(
    name="character",
    type="character",
    parameters = "vector",
    children = "vector",
    parent = "vector"
  )
)

##################################################################################################
###################################### SETTERS AND GETTERS #######################################
##################################################################################################
VariableNode$accessors(c("name","type","parameters","children"))

##################################################################################################
############################################# METHODS ############################################
##################################################################################################

##################################################################################################
# VariableNode$methods (
#   equals = function(var) {
#     identical(.self, var)
#   }
# )
##################################################################################################

# Checks if the type of the variable is continuous.
#
# Args:
#
# Returns:
#   A boolean indicating if the type of the variable is continuous.

VariableNode$methods (
  isContinuous = function() {
    return(type=="continuous")
  }
)

##################################################################################################

# Checks if the type of the variable is categorical.
#
# Args:
#
# Returns:
#   A boolean indicating if the type of the variable is categorical.

VariableNode$methods (
  isCategorical = function() {
    return(type=="categorical")
  }
)

##################################################################################################

# Gets the parameter located in a given position in the list of parameters.
#
# Args:
#   i: Position of the parameters.
#
# Returns:
#   The parameter in position i in the list of parameters.

VariableNode$methods (
  getParameterAt = function(i){
    parameters[[i]]
  }
)
##################################################################################################
# VariableNode$methods (
#   getChildren = function(){
#     return(children)
#   }
# )

##################################################################################################

# Computes the number of children.
#
# Args:
#
# Returns:
#   The number of children.

VariableNode$methods (
  numChildren = function() {
    return(length(getChildren()))
  }
)

##################################################################################################

# Computes the number of parameter nodes of this variable.
#
# Args:
#
# Returns:
#   The number of parameter nodes of this variable.

VariableNode$methods (
  numParameterNodes = function() {
    length(parameters)
  }
)

##################################################################################################

# Adds a child to this variable at the end of the list of children.
#
# Args:
#   child: The child to be added.
#
# Returns:
#   The updated variable.

VariableNode$methods (
  addChild = function(child){
    child$setParent(c(parent=.self))
    children <<- c(children,child)
  }
)

##################################################################################################

# Gets the child at a given position.
#
# Args:
#   i: Position of the child.
#
# Returns:
#   The child at position i.

VariableNode$methods (
  getChildAt = function(i){
    return(children[[i]])
  }
)

##################################################################################################

# Checks if this variable is a leaf in the tree, that is, it does not have children.
#
# Args:
#
# Returns:
#   A boolean indicating if this variable is a leaf.

VariableNode$methods (
  isLeaf = function() {
    return(length(getChildren())==0)
  }
)

##################################################################################################

# Removes the list of children from this variable.
#
# Args:
#
# Returns:
#   The updated variable.

VariableNode$methods (
  removeChildren = function() {
    # Firstly, the parent for each child is removed
    lapply(getChildren(), function(x) {x$setParent(list())})
    # Secondly, the children are removed from the list of children
    setChildren(list())
  }
)

##################################################################################################

# Removes the child at a position in the list of children.
#
# Args:
#   i: Position of the child to be removed.
#
# Returns:
#   The updated variable.

VariableNode$methods (
  removeChildAt = function(i) {
    # Firstly, the parent is removed
    getChildAt(i)$setParent(list())
    # Secondly, the children is removed from the list of children
    children <<- children[-c(i)]

  }
)

##################################################################################################

# Removes a child from the list of children.
#
# Args:
#   child: The child to be removed.
#
# Returns:
#   The updated variable.

VariableNode$methods (
  removeChild = function(child) {
    for(i in 1:numChildren()) {
      if(identical(child,getChildAt(i))) {
        removeChildAt(i)
        break
      }
    }
  }
)

##################################################################################################

# Sets a parent (must be unique) to this variable.
#
# Args:
#   p: The parent to be set.
#
# Returns:
#   The modified variable.

VariableNode$methods (
  setParent = function(p) {
    parent<<- c(p)
  }
)

##################################################################################################

# Checks if a variable has a parent.
#
# Args:
#
# Returns:
#   A boolean indicating if this variable has a parent.

VariableNode$methods (
  hasParent = function() {
    if (length(parent)>0)
      return(TRUE)
    return(FALSE)
  }
)

##################################################################################################

# Checks if a given parameter node belongs to this variable.
#
# Args:
#   p: The parameter to be checked.
#
# Returns:
#    A boolean indicating if the given parameter belongs to this variable.

VariableNode$methods (
  hasParameter = function(param) {
    for(i in 1:numParameterNodes()){
      if(identical(getParameterAt(i),param))
        return(TRUE)
    }
    return(FALSE)
  }
)

##################################################################################################

# Gets the parent of this variable.
#
# Args:
#
# Returns:
#    The parent of the current variable.

VariableNode$methods (
  getParent = function() {
    parent
  }
)

##################################################################################################

# Assigns a set of children to this variable.
#
# Args:
#   ch: A vector of child variables.
#
# Returns:
#    The modified variable.

VariableNode$methods (
  setChildren = function(ch){
    # First remove the current variable as the parent of all its children
    a = lapply(getChildren(), function(x) x$setParent(list()))
    children <<- ch
    a = lapply(getChildren(), function(x) x$setParent(.self))
    # a = ... es para evitar que imprima por pantalla al llamar al método. Hay alguna forma elegante de evitar esto?
  }
)

##################################################################################################

# Adds a parameter to this variable at the end of the list of parameters.
#
# Args:
#   param: The parameter to be added.
#
# Returns:
#   The updated variable.

VariableNode$methods (
  addParameter = function(param) {
    parameters <<- c(parameters, param)
  }
)

##################################################################################################

# Removes the list of parameters from this variable.
#
# Args:
#
# Returns:
#   The updated variable.

VariableNode$methods (
  removeParameters = function() {
    setParameters(list())
  }
)

##################################################################################################

# Removes the parameter at a position in the list of parameters.
#
# Args:
#   i: Position of the parameter to be removed.
#
# Returns:
#   The current variable is modified.

VariableNode$methods (
  removeParameterAt = function(i) {
    setParameters(getParameters()[-c(i)])
  }
)

##################################################################################################

# Removes a given parameter from the list of parameters.
#
# Args:
#   param: The parameter node to be removed.
#
# Returns:
#   The updated variable.

VariableNode$methods (
  removeParameter = function(param) {
    for(i in 1:numParameterNodes()) {
      if(identical(param,getParameterAt(i))) {
        removeParameterAt(i)
        break
      }
    }
  }
)

##################################################################################################

# Gets the position of a child given as parameter in the list of children.
#
# Args:
#   child: A child variable.
#
# Returns:
#   The position in which the child is stored in the list of children.

VariableNode$methods (
  getChildPosition = function(child) {
    for(i in 1:numChildren()){
      if(identical(child, getChildAt(i))){
        return(i)
      }
    }
    return(-1)
  }
)

##################################################################################################

# Computes the number of reached cases for each parameter node contained in this variable.
#
# Args:
#
# Returns:
#   A vector with the number of reached cases for each parameter node contained in this variable.

VariableNode$methods (
  numReachedCasesParameters = function() {
    vNumCases = c()
    for(i in 1:numParameterNodes()){
      vNumCases[i]=getParameterAt(i)$numReachedCases()
    }
    return(vNumCases)
  }
)

##################################################################################################

# Gets the names of the parameters contained in this variable.
#
# Args:
#
# Returns:
#   A vector with the names of the parameters contained in this variable.

VariableNode$methods (
  getParameterNames = function() {
    vNames=c()
    for(i in 1:numParameterNodes()){
      vNames[i]=getParameterAt(i)$getName()
    }
    return(vNames)
  }
)


##################################################################################################

# Gets the parameter given its name as argument.
#
# Args:
#   name: the name of the parameter.
#
# Returns:
#   The corresponding parameter.

VariableNode$methods (
  getParameter = function(name) {
    for(i in 1:numParameterNodes()){
      if(identical(getParameterAt(i)$getName(),name))
        return(getParameterAt(i))
    }
    return(NULL)
  }
)

##################################################################################################

# Collapses two given parameters of this variable by creating a new one:
# - The reached cases for the new parameter is the union of the former ones.
# - The potential for the new parameter is estimated.
# - The PDG structure above and bellow the two parameter nodes are modified.
# - The (number of) parameters of this variable are also modified.
#
# Args:
#   v: The first parameter node to be collapsed.
#   w: The second parameter node to be collapsed.
#
# Returns:
#   The modified PDG.

VariableNode$methods (
  collapseParameters = function(v,w, maxTerms=NULL) {

    var = v$getVar()
    #cat("\n       Cases before collapsing:", var$numReachedCasesParameters())

    if(!var$hasParameter(v) || !var$hasParameter(w)) stop("\n Parameter does not exist")
    if(identical(v,w)) stop("\n Identical parameters can't be collapsed")

    #newName = paste(v$getName(),w$getName(),sep="")
    #cat("\n       Creating new node:",newName )
    #vNew = CategoricalParameterNode$new(name=newName)

    if (.self$isCategorical()){
      vNew = CategoricalParameterNode$new()
    }
    else{
      vNew = ContinuousParameterNode$new(maxIntervals=v$getMaxIntervals())
    }

    vNew$getDistribution()$setVar(.self)

    #cat("\n       Merging cases from", v$getName(),"and",w$getName())
    vNew$getReachedData()$mergeCases(v$getReachedData())
    vNew$getReachedData()$mergeCases(w$getReachedData())

    for(pred in v$getPredecessors()){
      #cat("\n       For predecessor",pred$getName())

      pos = pred$getPositionsSucc(v)[1]
      #cat("\n           Successor position of", v$getName(), ":", pos)

      #cat("\n           Inserting", vNew$getName(),"in position", pos)
      pred$addSuccessorAt(vNew,pos)

      #cat("\n           Adding", pred$getName(),"as predecessor of", vNew$getName())
      vNew$addPredecessor(pred)

      #cat("\n           Removing", v$getName(), "as successor of", pred$getName())
      pred$removeSuccessor(v)

      #cat("\n           Removing", pred$getName(), "as predecessor of", v$getName())
      v$removePredecessor(pred)



    }

    for(pred in w$getPredecessors()){
      #cat("\n       For predecessor",pred$getName())

      pos = pred$getPositionsSucc(w)[1]
      #cat("\n           Successor position of", w$getName(), ":", pos)

      #cat("\n           Inserting", vNew$getName(),"in position", pos)
      pred$addSuccessorAt(vNew,pos)

      #cat("\n           Adding", pred$getName(),"as predecessor of", vNew$getName())
      vNew$addPredecessor(pred)

      #cat("\n           Removing", w$getName(), "as successor of", pred$getName())
      pred$removeSuccessor(w)

      #cat("\n           Removing", pred$getName(), "as predecessor of", w$getName())
      w$removePredecessor(pred)

    }

    #cat("\n      Removing", v$getName(), "from variable", var$getName())
    var$removeParameter(v)
    #cat("\n      Removing", w$getName(), "from variable", var$getName())
    var$removeParameter(w)
    #cat("\n      Adding", vNew$getName(), "to variable", var$getName())
    var$addParameter(vNew)

    #cat("\n       Re-learn the potential")
    #potNew = ProbabilityVector$new(var=var)
    #potNew$learning(vNew$getReachedData()$getColumnData(var))

    #cat("\n      Setting to", vNew$getName(),"the averaged potential of", vNew$getName(), "and", w$getName())

    #PARA CONTINUAS NO HACER LA MEDIA, SINO QUE APRENDEMOS DE NUEVO CON LOS DATOS FUSIONADOS.


    if(.self$isContinuous()){
      # potNew = MoTBFDensity$new(var=var, type=v$getDensityType())

      if(v$getDensityType()=="MOP"){
        potNew = MOP$new(var=var)
        potNew$fitMoP(vNew$getReachedData()$getColumnData(var),maxDegree=maxTerms)
      }
      else if(v$getDensityType()=="MTE") {
        potNew = MTE$new(var=var)
        potNew$fitMTE(vNew$getReachedData()$getColumnData(var), maxExpTerms=maxTerms)
      }
      else stop("Unknown type of density")


      #potNew$fitDensity(vNew$getReachedData()$getColumnData(var))
    }
    else{
      potNew = var$getAveragedPotential(v,w)
    }
    vNew$setDistribution(potNew)
    #cat("\n       Parameters:",getParameterNames())
    #cat("\n       Number of parameters:",numParameterNodes())
    #cat("\n       Cases after collapsing:", numReachedCasesParameters())
  }
)
##################################################################################################

# Creates a new parameter node from two parameters v and w. The new potential is computed as the
# average of the potentials of v and w. The computation of the average is weighted by the number of
# reached cases for each parameter node. This method works for both categorical and continuous
# potentials.
#
# Args:
#   v: A parameter node.
#   w: A parameter node.
#
# Returns:
#   An averaged parameter node of v and w.
#

VariableNode$methods (
  getAveragedPotential = function(v,w) {

    if(.self$isCategorical())
      potNew = ProbabilityVector$new(var=.self)
    else{
      if (v$getDensityType()=="MOP") potNew = MOP$new(var=.self)
      else      		               potNew = MTE$new(var=.self)

    }

    paramsV = v$getDistribution()$getParameters()
    paramsW = w$getDistribution()$getParameters()

    a = v$numReachedCases()
    b = w$numReachedCases()

    # if there is no reached cases in v and w means that we cannot average the potentials
    # since the denominator (a+b)==0
    if(a==0 && b==0){
      potNew$setParameters(paramsV) # or paramsW, they are equals
      return(potNew)
    }



    #---------------------------------------------------------------------------------------------
    # If the variable is continuous, the number of parameters of both potentials may be different.
    # So, one of the parameter vector is filled with 0s.
    if(.self$isContinuous()) {

      nParamsV = v$getDistribution()$numParameters()
      nParamsW = w$getDistribution()$numParameters()
      numZeros = abs(nParamsV-nParamsW)
      vZeros = rep(0,numZeros)
      if(nParamsV>nParamsW){
        paramsW = c(paramsW,vZeros)
      }
      else {
        paramsV = c(paramsV,vZeros)
      }
    }
    #---------------------------------------------------------------------------------------------

    newParams = (paramsV*a + paramsW*b)/(a+b)

    #potNew = ProbabilityVector$new(var=.self)

    potNew$setParameters(newParams)
    return(potNew)
  }
)

###################################################################################################

##################################################################################################
######################################## CLASS DEFINITION ########################################
##################################################################################################

CategoricalVariableNode <- setRefClass(
  Class = "CategoricalVariableNode",
  fields = list(
    states="character"
  ),
  contains="VariableNode"
)
##################################################################################################
###################################### SETTERS AND GETTERS #######################################
##################################################################################################
CategoricalVariableNode$accessors(c("states"))

##################################################################################################
########################################### METHODS ##############################################
##################################################################################################

##################################################################################################

# Initializes the categorical variable with a its name and states. All these
# arguments are mandatory.
#
# Args:
#   name: The name of the continuous variable.
#   states: a vector of strings with the names of the variable states.
#
# Returns:
#   The initialized CategoricalVariableNode object.

CategoricalVariableNode$methods (
  initialize = function(name, states){
    setName(name)
    setStates(states)
    setType("categorical")
    setParent(list())
  }
)

##################################################################################################

# Shows relevant information about the variable: name and states.
#
# Args:
#
# Returns:
#   No arguments are returned. Only the information is displayed on the screen.

CategoricalVariableNode$methods (
  show = function() {
    cat("  Name: ", getName())
    cat("\n  Type: ", getType())
    cat("\nStates: ", getStates())
    cat("\n")
  }
)

##################################################################################################

# Gets the state at a position.
#
# Args:
#   i: position of the state.
#
# Returns:
#   A String indicating the state at position i.

CategoricalVariableNode$methods (
  getState = function(i) {
    getStates()[i]
  }
)

##################################################################################################

# Computes the number of states of this variable.
#
# Args:
#
# Returns:
#   The number of states of this variable.

CategoricalVariableNode$methods (
  numStates = function() {
    length(getStates())
  }
)

##################################################################################################

##################################################################################################
######################################## CLASS DEFINITION ########################################
##################################################################################################
ContinuousVariableNode <- setRefClass(
  Class = "ContinuousVariableNode",
  fields = list(
    min="numeric",
    max="numeric"
  ),
  contains="VariableNode"
)

##################################################################################################
###################################### SETTERS AND GETTERS #######################################
##################################################################################################
ContinuousVariableNode$accessors(c("min", "max"))

##################################################################################################
########################################### METHODS ##############################################
##################################################################################################

##################################################################################################

# Initializes the continuous variable with a its name, minimum and maximum value. All these
# arguments are mandatory.
#
# Args:
#   name: The name of the continuous variable.
#   min: The minimum value of the variable from the database used for learning.
#   max: The maximum value of the variable from the database used for learning.
#
# Returns:
#   The initialized ContinuousVariableNode object.

ContinuousVariableNode$methods (
  initialize = function(name, min, max){
    setName(name)
    setMin(min)
    setMax(max)
    setType("continuous")
    setParent(list())
  }
)

##################################################################################################

# Shows relevant information about the variable: name, type, minimum and maximum.
#
# Args:
#
# Returns:
#   No arguments are returned. Only the information is displayed on the screen.

ContinuousVariableNode$methods (
  show = function() {
    cat("  Name: ", getName())
    cat("\n  Type: ", getType())
    cat("\n   Min: ", getMin())
    cat("\n   Max: ", getMax())
  }
)

##################################################################################################


##################################################################################################
######################################## CLASS DEFINITION ########################################
##################################################################################################
UnivariatePotential <- setRefClass(
  Class = "UnivariatePotential",
  fields = list(
    parameters = "numeric",
    var = "VariableNode"
  )
)

##################################################################################################
###################################### SETTERS AND GETTERS #######################################
##################################################################################################
UnivariatePotential$accessors(c("parameters","var"))

##################################################################################################
########################################### METHODS ##############################################
##################################################################################################

##################################################################################################

# Gets the number of parameters included in the potential.
#
# Args:
#
# Returns:
#   The number of parameters of the potential.

UnivariatePotential$methods (
  numParameters = function() {
    length(getParameters())
  }
)
##################################################################################################

# Gets the parameter at a given position.
#
# Args:
#
# Returns:
#   The parameter at position i.

UnivariatePotential$methods (
  getParameterAt = function(i) {
    parameters[i]
  }
)
##################################################################################################
##################################################################################################
######################################## CLASS DEFINITION ########################################
##################################################################################################
MoTBFDensity <- setRefClass(
  Class = "MoTBFDensity",
  fields = list(
    type="character"
  ),
  contains="UnivariatePotential"
)

##################################################################################################
###################################### SETTERS AND GETTERS #######################################
##################################################################################################
MoTBFDensity$accessors(c("type"))

##################################################################################################
########################################### METHODS ##############################################
##################################################################################################

# Converts this density into a String with function format. Note that this method is not
# implemented in this class, but in the subclasses.
#
# Args:
#
# Returns:
#   A String representing the density.

MoTBFDensity$methods (
  ToString = function(){}
)

##################################################################################################

# Converts the String that represents this density into a Function that can be evaluated.
#
# Args:
#   x: The generic variable of a function.
#
# Returns:
#   The corresponding Function of this density.

MoTBFDensity$methods (
  asFunction = function(x) {
    as.function(alist(x = , eval(parse(text=ToString()))))
  }
)

##################################################################################################

# Plots de density function into a file.
#
# Args:
#   pdfFile: The file name in which the plot will be saved.
#   nameParameter: Name of the parameter to be included in the title of the plot.
#
# Returns:
#   No arguments are returned. Only the file with the plot is saved.

MoTBFDensity$methods (
  plot = function(pdfFile, nameParameter) {
    if(length(dev.list())!=0) dev.off() # reboot the graphical window
    pdf(pdfFile)
    nameVar=getVar()$getName()
    min = getVar()$getMin()
    max = getVar()$getMax()
    sizeReachedData = getVar()$getParameter(nameParameter)$numReachedCases()

    hist(getVar()$getParameter(nameParameter)$getReachedData()$getColumnData(getVar()),
         xlab=nameVar, ylab=paste("f(",nameVar,")",sep=""),
         xlim = c(min,max), freq=FALSE,
         main=paste("Parameter: ",nameParameter," - Reached cases: ",sizeReachedData,"\n",
                    getType()," - Number of parameters: ",numParameters(),sep=""))
    curve(asFunction()(x),add=T, col="blue")


    dev.off()
  }
)

##################################################################################################

# Computes the definite integral of the density function in the range [min(var),max(var)].
#
# Args:
#
# Returns:
#   The value of the integral.

MoTBFDensity$methods (
  definiteIntegral= function() {
    integrate(asFunction(),getVar()$getMin(),getVar()$getMax())$value
  }
)

##################################################################################################

# Normalizes the parameters of the density to have a function that integrates up one.
#
# Args:
#
# Returns:
#   The parameters of the density are updated.

MoTBFDensity$methods (
  normalize = function () {
    setParameters(parameters/definiteIntegrate())
  }
)

##################################################################################################

# Computes the BIC score of this density for a given data
#
# Args:
#
# Returns:
#   The BIC score.

MoTBFDensity$methods (
  getBIC=function(data){
    pPx = asFunction()(data)
    sum(log(pPx))-(1/2*(length(parameters))*log(length(data)))
  }
)

##################################################################################################

# Evaluate a given value into this density function.
#
# Args:
#   value: A continuous value
#
# Returns:
#   The value of the density for the given argument.

MoTBFDensity$methods (
  evaluate=function(value){
    asFunction()(value)
  }
)
##################################################################################################

# Fits a MoTBF density ("MOP" or "MTE") from data.
#
# Args:
#   data: A univariate vector of data.
#   nterms: For "MOP" represents the degree of the polynomial (number of parameters = 1+nterms).
#           For "MTE" is the number of exponential terms (number of parameters = 1+nterms). This
#           argument is not mandatory. If it is not given, the optimal number of parameters is
#           computed using a BIC-based optimisation procedure.
#
# Returns:
#   A MoTBF density.

MoTBFDensity$methods (
  fitDensity=function(data, nTerms=NULL, maxTerms=NULL){

    switch(getType(),
           MOP={ pot = MOP$new(var=getVar())
           pot$fitMoP(data,maxDegree=maxTerms)},

           MTE={
             pot = MTE$new(var=getVar())
             pot$fitMTE(data,maxExpTerms=maxTerms)
           },

           { stop("Unknown density type.") }
    )
    return(pot)
  }
)
##################################################################################################

# Simulates a value from an univariate MoTBF density ("MOP" or "MTE").
#
# Args:
#
# Returns:
#   The simulated value.

MoTBFDensity$methods (
  simulateValue = function(){}
)

##################################################################################################
######################################## CLASS DEFINITION ########################################
##################################################################################################
ProbabilityVector <- setRefClass("ProbabilityVector",
                                 fields = list(
                                 ),
                                 contains="UnivariatePotential"
)

##################################################################################################
###################################### SETTERS AND GETTERS #######################################
##################################################################################################

##################################################################################################
########################################### METHODS ##############################################
##################################################################################################

##################################################################################################

# Initializes the object.
#
# Args:
#   var: The variable being part of the potential.
#   parameters: The probabilities of the potential (non mandatory argument).
#
# Returns:
#   The initialized object.

ProbabilityVector$methods (
  initialize = function(var, parameters = NULL) {
    if (var$isCategorical()){
      var <<- var
      if (is.null(parameters)){
        parameters <<- rep(1/var$numStates(), var$numStates())
      }
      else {
        parameters <<- parameters
      }
    }
    else {
      stop("Variable must be categorical")
    }
  }
)

##################################################################################################

# Shows relevant information about the object: variable, states and parameters.
#
# Args:
#
# Returns:
#   No arguments are returned. Only the information is displayed on the screen.

ProbabilityVector$methods (
  show = function() {
    cat("     Variable: "); cat(getVar()$getName())
    cat("\n       States: "); cat(getVar()$getStates())
    cat("\n   Parameters: "); cat(parameters)
  }
)

##################################################################################################

# Learns the probabilities of the potential.
#
# Args:
#   data: A vector of values for the corresponding variable.
#
# Returns:
#   The object with the learned parameters.

ProbabilityVector$methods(
  learning = function(data) {
    numStates = getVar()$numStates()
    n = length(data)
    # Se puede aplicar algun calculo vectorial aqui?????
    for(i in 1:numStates) {
      # máxima verosimilitud con Laplace correction
      prob=(length(data[data==getVar()$getState(i)])+1)/(n+numStates)
      parameters[i] <<- prob
    }
  }
)

##################################################################################

# Simulates a set of values from a probability table.
#
# Args:
#
# Returns:
#   The vector of simulated labels.

ProbabilityVector$methods (
  simulateValue = function(){

    v = cumsum(getParameters())
    x = runif(1)
    i=1
    while (x>v[i]) i=i+1
    state = getVar()$getState(i)
    return(state)
  }
)



##################################################################################################

# Creates a bar plot representing the probabilities of the potential.
#
# Args:
#   pdfFile: The file name in which the plot will be saved.
#   nameParameter: Name of the parameter to be included in the title of the plot.
#
# Returns:
#   No arguments are returned. Only the file with the plot is saved.

ProbabilityVector$methods (
  plot = function(pdfFile,nameParameter){
    if(length(dev.list())!=0) dev.off() # reboot the graphical window
    pdf(pdfFile)
    barX <- barplot(getParameters(),ylim=c(0,1.1), cex.main=1,
                    ylab=paste("P(", getVar()$getName(),")",sep=""),
                    main=nameParameter, names.arg=getVar()$getStates())
    text(x=barX, y=getParameters()+0.04, round(getParameters(),4))
    dev.off()
  }
)

##################################################################################################

##################################################################################################
######################################## CLASS DEFINITION ########################################
##################################################################################################
MOP <- setRefClass(
  Class = "MOP",
  fields = list(
  ),
  contains="MoTBFDensity"
)

##################################################################################################
########################################### METHODS ##############################################
##################################################################################################

##################################################################################################

# Initializes the MOP with the variable and its type.
#
# Args:
#   var: The variable of this univariate density.
#
# Returns:
#   The initialized MOP object.

MOP$methods (
  initialize = function (var) {
    setVar(var)
    setType("MOP")
  }
)

##################################################################################################

# Gets the degree of the MOP density.
#
# Args:
#
# Returns:
#   The degree of the density.

MOP$methods (
  degree = function() {
    length(parameters)-1
  }
)

##################################################################################################

# Converts this MOP density into a String with function format.
#
# Args:
#
# Returns:
#   A String representing the density.

MOP$methods (
  ToString = function () {
    sign = parameters; sign[sign<0]=""; sign[sign>=0]="+"
    str = paste(parameters[1], sep="")
    if(degree()>0){
      for(i in 1:(length(parameters)-1)){
        if(i==1) str = paste(str,sign[i+1], parameters[i+1], "*x", sep="")
        else     str = paste(str,sign[i+1], parameters[i+1], "*x^",i, sep="")
      }
    }
    else {
      str = paste(str,"+0*x", sep="") # Only when the MOP is a constant
    }
    str
  }
)

##################################################################################################

# Shows relevant information about the density: variable and density function.
#
# Args:
#
# Returns:
#   No arguments are returned. Only the information is displayed on the screen.

MOP$methods (
  show = function() {
    if(degree()>-1) {
      cat("     Variable: "); cat(getVar()$getName(),"\n")
      cat("      Density:",ToString())
    }
  }
)

##################################################################################################

# Fits a MOP density with a fixed degree to a given data.
#
# Args:
#   data: A Vector of continuous values.
#   degree: A Integer indicating the degree of the fitted density.
#
# Returns:
#   A MOP density

MOP$methods (
  fitMoP.FixedDegree=function(data,degree){

    a = getVar()$getMin()
    b = getVar()$getMax()

    if(length(unique(data))<=(degree+1)){
      setParameters(c(1/(b-a))) # UNIFORM
    }
    else {

      nterms = degree + 1
      data = sort(data)
      Fx = ecdf(data)
      y = unique(Fx(data))
      data = unique(data)

      n = length(data)
      nPoints = 10^(-3)
      Dmat=rep(1,n)
      for(i in 1:nterms){
        Dmat=cbind(Dmat,cbind(data^i))
      }
      dvec= t(Dmat)%*%y
      Dmat = t(Dmat)%*%Dmat

      xPoints=seq(a, b, nPoints);
      if (xPoints[length(xPoints)]!=b)
        xPoints[length(xPoints)] = b # Ensure that b is in the sequence


      m1=0
      for(j in 1:nterms){
        m1=rbind(m1,j*xPoints^(j-1))
      }

      m2=1; m3=1
      for(i in 1:nterms){
        m2=rbind(m2,a^i)
        m3=rbind(m3,b^i)
      }

      Amat=cbind(m2, m3, m1)
      bvec= c(0, 1, rep(nPoints,ncol(m1)))

      #tr=solve.QP(Dmat, dvec, Amat, bvec, meq=2)



      tr=tryCatch(solve.QP(Dmat, dvec, Amat, bvec, meq=2), error = function(e) NULL)

      if(is.null(tr)){
        return(NULL)
        #options (digits=22)
        #   	cat("\n\n\nDegree:", degree)
        #	cat("\na=",a)
        #	cat("\nb=",b,"\n")
        #		print(data)
        #		cat("\nLength(unique(data)):",length(unique(data)))
        #		hist(data, xlim=c(a,b))
        #		stop(1)
      }

      Px=deriv(polynomial(tr$solution))

      setParameters(Px[1:length(Px)])
    }
  }
)

##################################################################################

# Fits a MOP density to a given data. The parameter degree is optional. If it is
# not given, the optimal degree is computed.
#
# Args:
#   data: A Vector of continuous values.
#   degree: The degree of the fitted density (non mandatory argument).
#   maxDegree: The maximum degree to be explored.
#
# Returns:
#   A MOP density

MOP$methods (
  fitMoP=function(data, maxDegree = NULL){

    degree=0
    newBIC=Inf
    BIC=-Inf
    vBIC=c()
    notImprove = 0
    bestParams = c(0)
    repeat {

      if(notImprove==2 || degree>maxDegree) {break}

      result = .self$fitMoP.FixedDegree(data, degree)

      if(is.null(result)) break

      newBIC=getBIC(data)

      vBIC=c(vBIC,newBIC)

      if(newBIC>BIC){
        bestParams = getParameters()
        BIC=newBIC
        notImprove=0
      }
      else{
        notImprove=notImprove+1
      }
      degree=degree+1
    }
    setParameters(bestParams[1:length(bestParams)])
  }
)

##################################################################################

# Computes the symbolic integral of an MOP with format:
# f(x)= a+b*x+c*x^2 + ...
#
# Args:
#
# Returns:
#   A vector with the coefficients a,b,c ... of the integral:
#             f(x)= a+b*x+c*x^2 + ...


MOP$methods (
  symbolicIntegral = function(){
    coef(integral(polynomial(getParameters())))
  }
)

##################################################################################

# Simulates a set of values from an univariate MOP density.
#
# Args:
#
# Returns:
#   The vector of simulated values.

MOP$methods (
  simulateValue = function(){

    coeffCDF = .self$symbolicIntegral()

    pol = polynomial(c(coeffCDF))
    f = as.function(pol)
    min = getVar()$getMin(); max = getVar()$getMax()

    # The minimum y value of CDF is computed to move the CDF to the y-range [0,1]
    yScaleValue = f(min)

    pol = pol-yScaleValue-runif(1)

    f = as.function(pol)
    # We create a function from the new string after substracting yScaleValue and runif(1)
    root=uniroot(f, c(min,max))$root
    return(root)
  }
)

##################################################################################

##################################################################################################
######################################## CLASS DEFINITION ########################################
##################################################################################################
MTE <- setRefClass(
  Class = "MTE",
  fields = list(
  ),
  contains="MoTBFDensity"
)
##################################################################################################
########################################### METHODS ##############################################
##################################################################################################

##################################################################################################

# Initializes the MTE with the variable and its type.
#
# Args:
#   var: The variable of this univariate density.
#
# Returns:
#   The initialized MTE object.

MTE$methods (
  initialize = function (var) {
    setVar(var)
    setType("MTE")
  }
)

##################################################################################################

# Gets the number of exponential terms of the MTE density.
#
# Args:
#
# Returns:
#   The number of exponential terms.

MTE$methods (
  numExpTerms = function() {
    numParameters()-1
  }
)
##################################################################################################

# Converts this MTE density into a String with function format.
#
# Args:
#
# Returns:
#   A String representing the density.

MTE$methods (
  ToString = function() {
    str = parameters[1]
    sign = parameters; sign[sign<0]=""; sign[sign>=0]="+"

    if(numExpTerms()>0) {
      for(i in 2: numParameters()){
        if(i<4) str = paste(str,sign[i], parameters[i], "*exp(",ifelse(i%%2==1, "-", ""),"x)", sep="")
        else    str = paste(str,sign[i], parameters[i], "*exp(",ifelse(i%%2==1, "-", ""),i%/%2,"*x)", sep="")
      }
    }
    else {
      str = paste(str,"+0*x", sep="") # Only when the MTE is a constant
    }
    return(str)
  }
)

##################################################################################################

# Shows relevant information about the density: variable and density function.
#
# Args:
#
# Returns:
#   No arguments are returned. Only the information is displayed on the screen.

MTE$methods (
  show = function() {
    cat("Variable: "); cat(getVar()$getName(),"\n")
    cat(ToString())
  }
)

##################################################################################################

# Fits a MTE density with a fixed number of terms to a given data.
#
# Args:
#   data: A Vector of continuous values.
#   nExpTerms: A Integer indicating the number of exponential terms.
#
# Returns:
#   A MTE density with nExpTerms exponential terms


MTE$methods (
  fitMTE.FixedNumTerms=function(data, nExpTerms){

    options(digits=22)
    if ((nExpTerms%%2)!=0) { stop("The number of exponential terms must be even.")}

    a = getVar()$getMin()
    b = getVar()$getMax()

    if(length(unique(data))<=(nExpTerms+1)*2){
      setParameters(c(1/(b-a))) # UNIFORM
    }
    else {


      nPairedTerms = nExpTerms/2

      data=sort(data)
      Fx = ecdf(data)
      y = unique(Fx(data))
      data = unique(data)

      n=length(data)

      Dmat=c(data)#introducimos la x en el termino independiente para que al derivar se mantenga
      for(i in 1:nPairedTerms){
        Dmat=rbind(Dmat,exp((i)*data), exp((-i)*data))
      }
      dvec = Dmat%*%y
      Dmat = Dmat%*%t(Dmat)

      nPoints = 10^(-3)
      xPoints=seq(a, b, nPoints) # Shouldn't be nPoints depends on the range width?

      if (xPoints[length(xPoints)]!=b)
        xPoints[length(xPoints)] = b # Ensure that b is in the sequence

      t=1
      for(i in 1:nPairedTerms){
        t=rbind(t,(i)*exp( (i)*xPoints),
                (-i)*exp((-i)*xPoints))
      }

      m1=matrix(t, nrow=2*nPairedTerms+1)

      m2=c(a); m3=c(b)
      for(i in 1:nPairedTerms){
        m2=rbind(m2, exp((i)*a), exp((-i)*a))
        m3=rbind(m3, exp((i)*b), exp((-i)*b))
      }
      Amat=cbind(m2, m3, m1)
      bvec= c(0,1, rep(nPoints, length(xPoints)))


      tr=tryCatch(solve.QP(Dmat, dvec, Amat, bvec, meq=2), error = function(e) NULL)

      if(is.null(tr)){
        cat("\n\n\nExpTerms:", nExpTerms)
        cat("\na=",a)
        cat("\nb=",b,"\n")
        print(data)
        hist(data, xlim=c(a,b))
        stop(1)
      }
      parametros = tr$solution
      #parametros = solve.QP(Dmat, dvec, Amat, bvec, meq=2)$solution





      ######################## Derive #######################
      n = length(parametros)

      npar=parametros[(1:n)%%2==0]

      par11=c()
      for(i in 1:length(npar)){
        par11=cbind(par11,(i)*npar[i])
      }

      nimp=parametros[(1:n)%%2==1]

      par22=c()
      for(i in 1:(length(nimp)-1)){
        par22=cbind(par22, (-i)*nimp[i+1])
      }

      h=c(parametros[1])
      for(i in 1:length(npar)){
        h=c(h, par11[i], par22[i])
      }
      setParameters(h)
    }
  }
)

##################################################################################

# Fits a MTE density to a given data. The parameters nPairedTerms and maxTerms are
# optional. If they are not given, the optimal number of paired terms is computed.
#
# Args:
#   data: A Vector of continuous values.
#   nPairedTerms: A Integer indicating the number of pairs of exponential terms (non mandatory argument).
#   maxTerms: The maximum number of terms to be explored
#
# Returns:
#   A MTE density

MTE$methods (
  fitMTE=function(data, maxExpTerms = NULL){



    if(!is.null(maxExpTerms) && (maxExpTerms%%2)!=0) { stop("The maximum number of exponential terms must be even.")}

    nExpTerms=2
    newBIC=Inf
    BIC=-Inf
    vBIC=c()
    notImprove = 0
    bestParams = c(0)

    repeat {

      if(notImprove==2 || nExpTerms>maxExpTerms) {break}

      .self$fitMTE.FixedNumTerms(data, nExpTerms)

      newBIC=getBIC(data)

      vBIC=c(vBIC,newBIC)

      if(newBIC>BIC){
        bestParams = getParameters()
        BIC=newBIC
        notImprove=0
      }
      else{
        notImprove=notImprove+1
      }
      nExpTerms=nExpTerms+2
    }
    setParameters(bestParams[1:length(bestParams)])

  }
)



##################################################################################

# Computes the symbolic integral of an MTE with format:
# f(x)= a+b*exp(x/5)+c*exp(-x/5)+d*exp(2*x/5)+f*exp(-2*x/5) + ...
#
# Args:
#
# Returns:
#   A vector with the coefficients a,b,c,d,f, ... of the integral:
#         a*x + b*exp(x/5)+c*exp(-x/5) + d*exp(2*x/5) + f*exp(-2*x/5) + ...

MTE$methods (
  symbolicIntegral = function(){
    v = getParameterAt(1)
    if((numParameters())>1) {
      for(i in 2:numParameters()){
        numerator = i%/%2
        if(i%%2==1) numerator = -numerator
        v = c(v,getParameterAt(i)*(5/numerator))
      }
    }
    return(v)
  }
)

##################################################################################

# Simulates a set of values from an univariate MTE density.
#
# Args:
#
# Returns:
#   The vector of simulated values.

MTE$methods (
  simulateValue = function(){
    coeffCDF = .self$symbolicIntegral()

    # Create an string representing the symbolic integral above
    # Se puede poner aparte en una función
    ############################################################################
    str = paste(coeffCDF[1], "*x", sep="")
    if(numParameters()>1) {
      for(i in 2:numParameters()){
        numerator = i%/%2
        if(i%%2==1) numerator = -numerator
        if(coeffCDF[i]>=0) sign="+" else sign=""
        str = paste(str,sign, coeffCDF[i], "*exp(",numerator,"*x)", sep="")
      }
    }
    ############################################################################
    # We create a function from the string above
    f = as.function(alist(x = , eval(parse(text=str))))
    min = getVar()$getMin()
    max = getVar()$getMax()
    # The minimum y value of CDF is computed to move the CDF to the y-range [0,1]
    yScaleValue = f(min)
    str = paste(-yScaleValue-runif(1), "+", str, sep="")
    # We create a function from the new string after substracting yScaleValue and runif(1)
    f = as.function(alist(x = , eval(parse(text=str))))
    root=uniroot(f, c(min,max))$root
    return(root)
  }
)








# #########################################################################################################
# MTE$methods (
#   combine = function(mte) {
#
#     f = getParameters()
#     g = mte$getParameters()
#
#     nPairedTerms= (length(f)-1)/2 + (length(g)-1)/2
#     nParameters = nPairedTerms*2+1
#
#     sign = c(0)
#     for (i in 1:nPairedTerms){
#       sign = c(sign,i,-i)
#     }
#
#     factors = rep(0,nParameters)
#     for(i in 1:length(f)){
#       for(j in 1:length(g)) {
#         result = f[i]*g[j]
#         pos = which(sign[i]+sign[j]==sign)
#         factors[pos]=factors[pos]+result
#       }
#     }
#     setParameters(factors)
#   }
# )
# #########################################################################################################





##################################################################################################
######################################## CLASS DEFINITION ########################################
##################################################################################################
ParameterNode <- setRefClass(
  Class = "ParameterNode",
  fields = list(
    name="character",
    predecessors = "list",
    successors = "list",
    distribution = "UnivariatePotential",
    reachedData = "Database"
  )
)

##################################################################################################
###################################### SETTERS AND GETTERS #######################################
##################################################################################################
ParameterNode$accessors(c("name","predecessors", "successors", "distribution", "reachedData"))

##################################################################################################
########################################### METHODS ##############################################
##################################################################################################

##################################################################################################

# Plots the potential into a file.
#
# Args:
#   pdfFile: The file name in which the plot will be saved.
#
# Returns:
#   No arguments are returned. Only the file with the plot is saved.

ParameterNode$methods (
  plot = function(pdfFile) {
    getDistribution()$plot(pdfFile, getName())
  }
)

##################################################################################################

# Gets the type of this parameter node.
#
# Args:
#
# Returns:
#   A String indicating the type of the parameter ("categorical" or "continuous").

ParameterNode$methods (
  getType = function() {
    getDistribution()$getVar()$getType()
  }
)


##################################################################################################

# Checks if the type of the parameter is continuous.
#
# Args:
#
# Returns:
#   A boolean indicating if the type of the parameter is continuous.

ParameterNode$methods (
  isContinuous = function() {
    getType()=="continuous"
  }
)

##################################################################################################

# Checks if the type of the parameter is categorical.
#
# Args:
#
# Returns:
#   A boolean indicating if the type of the parameter is categorical.

ParameterNode$methods (
  isCategorical = function() {
    getType()=="categorical"
  }
)

##################################################################################################

# Gets the successor at a position.
#
# Args:
#   i: Position of the successor.
#
# Returns:
#   The successor at position i.

ParameterNode$methods (
  getSuccessorAt = function(i) {
    successors[[i]]
  }
)

##################################################################################################

# Gets the predecessor at a position.
#
# Args:
#   i: Position of the predecessor.
#
# Returns:
#   The predecessor at position i.

ParameterNode$methods (
  getPredecessorAt = function(i) {
    predecessors[[i]]
  }
)
##################################################################################################

# Computes the number of successors.
#
# Args:
#
# Returns:
#   The number of successors.

ParameterNode$methods (
  numSuccessors = function() {
    length(successors)
  }
)

##################################################################################################

# Computes the number of predecessors.
#
# Args:
#
# Returns:
#   The number of predecessors.

ParameterNode$methods (
  numPredecessors = function() {
    length(predecessors)
  }
)

##################################################################################################

# Check if a successor is included in the list of successors of this parameter node.
#
# Args:
#   succ: The successor to be checked.
#
# Returns:
#   A boolean indicating if succ is a successor of this parameter node.

ParameterNode$methods (
  isSuccessor = function (succ){
    any(lapply(getSuccessors(), function(x) identical(x,succ))==TRUE)
  }
)

##################################################################################################

# Check if a predecessor is included in the list of predecessors of this parameter node.
#
# Args:
#   pred: The predecessor to be checked.
#
# Returns:
#   A boolean indicating if pred is a predecessor of this parameter node.

ParameterNode$methods (
  isPredecessor = function (pred){
    any(lapply(getPredecessors(), function(x) identical(x,pred))==TRUE)
  }
)
##################################################################################################

# Adds a successor to this parameter node at the end of the list of successors.
#
# Args:
#   succ: The successor to be added.
#
# Returns:
#   The updated parameter node.

ParameterNode$methods (
  addSuccessor = function(succ) {
    #if(!isSuccessor(succ)){ Puede existir dos succesores que sean el mismo!!!!!!
    successors <<- c(successors, succ)
    # }
  }
)

##################################################################################################

# Adds a successor to this parameter node at a position. The successors on the right are moved
# one position
#
# Args:
#   succ: The successor to be added.
#   i: The position in which the new successor will be located.
#
# Returns:
#   The updated parameter node.

ParameterNode$methods (
  addSuccessorAt = function(succ,i) {
    successors <<- append(successors, succ, i-1)
  }
)

##################################################################################################

# Sets a successor to this parameter node at a given position. The former successor is overwritten.
#
# Args:
#   succ: The successor to be set.
#   i: The position in which the new successor will be located.
#
# Returns:
#   The updated parameter node.

ParameterNode$methods (
  setSuccessorAt = function(succ,i) {
    successors[[i]] <<- succ
  }
)

##################################################################################################

# Adds a predecessor to this parameter node at the end of the list of predecessors.
#
# Args:
#   pred: The predecessor to be added.
#
# Returns:
#   The updated parameter node.

ParameterNode$methods (
  addPredecessor = function(pred) {
    # if(!isPredecessor(pred)){
    predecessors <<- c(predecessors, pred)
    #  }
  }
)

##################################################################################################

# Gets the variable to which this parameter node belongs.
#
# Args:
#
# Returns:
#   The variable to which this parameter node belongs.

ParameterNode$methods (
  getVar = function() {
    getDistribution()$getVar()
  }
)

##################################################################################################

# Checks if this parameter node is a leaf in the tree, that is, it does not have successors.
#
# Args:
#
# Returns:
#   A boolean indicating if this parameter node is a leaf.

ParameterNode$methods (
  isLeaf = function() {
    return(numSuccessors()==0)
  }
)

##################################################################################################

# Removes the list of predecessors from this parameter node.
#
# Args:
#
# Returns:
#   The updated parameter node.

ParameterNode$methods (
  removePredecessors = function() {
    setPredecessors(list())
  }
)

##################################################################################################

# Removes the list of successors from this parameter node.
#
# Args:
#
# Returns:
#   The updated parameter node.

ParameterNode$methods (
  removeSuccessors = function() {
    setSuccessors(list())
  }
)

##################################################################################################

# Removes the successor at a position in the list of successors.
#
# Args:
#   i: Position of the successor to be removed.
#
# Returns:
#   The updated parameter node.

ParameterNode$methods (
  removeSuccessorAt = function(i) {
    setSuccessors(getSuccessors()[-c(i)])
  }
)

##################################################################################################

# Removes a successor from the list of successors.
#
# Args:
#   succ: The successor to be removed.
#
# Returns:
#   The updated parameter node.

ParameterNode$methods (
  removeSuccessor = function(succ) {
    if(numSuccessors()>0) {
      for(i in 1:numSuccessors()) {
        if(identical(succ,getSuccessorAt(i))) {
          removeSuccessorAt(i)
          break
        }
      }
    }
  }
)

##################################################################################################

# Removes the predecessor at a position in the list of predecessors.
#
# Args:
#   i: Position of the predecessor to be removed.
#
# Returns:
#   The updated parameter snode.

ParameterNode$methods (
  removePredecessorAt = function(i) {
    setPredecessors(getPredecessors()[-c(i)])
  }
)

##################################################################################################

# Removes a predecessor from the list of predecessors.
#
# Args:
#   pred: The predecessor to be removed.
#
# Returns:
#   The updated parameter node.

ParameterNode$methods (
  removePredecessor = function(pred) {
    if(numPredecessors()>0) {
      for(i in 1:numPredecessors()) {
        if(identical(pred,getPredecessorAt(i))) {
          removePredecessorAt(i)
          break
        }
      }
    }
  }
)

##################################################################################################

# Returns a list of successors of this parameter node that belong to a given variable.
#
# Args:
#   var: The variable to which the returned successors belong.
#
# Returns:
#   A list of successors included in var.

ParameterNode$methods (
  getSuccessorsVariable = function(var) {
    vIndexes = c()
    for(i in 1:numSuccessors()) {
      if(identical(getSuccessorAt(i)$getVar(),var)) {
        vIndexes = c(vIndexes,i)
      }
    }
    return(getSuccessors()[vIndexes])
  }
)

##################################################################################################

# Returns the number of successors included in a given variable.
#
# Args:
#   var: A variable
#
# Returns:
#   The number of successors satisfying the condition.

ParameterNode$methods (
  numSuccessorsVariable = function(var) {
    length(getSuccessorsVariable(var))
  }
)

##################################################################################################

# Returns the unique successor included in a given variable and with a given label (index):
# succ(v, variable, index)
#
# Args:
#   var: A variable
#   index: The position of the successor in the list of successors included in var.
#
# Returns:
#   The unique successor satisfying the condition.

ParameterNode$methods (
  succ = function(variable,index) {
    if (numSuccessors()>0)
      getSuccessorsVariable(variable)[[index]]
  }
)

##################################################################################################

# Returns the parameter nodes reached by a given assignment bellow a given parameter node.
#
# Args:
#   node: A parameter node.
#   assignment: An assigment.
#   reachedNodes: Argument used in the recursivity to store the final list of reached nodes.
#
# Returns:
#   The parameter nodes reached by the assignment.

ParameterNode$methods (
  reachedNodesFromNode = function(node, assignment, reachedNodes) {

    #cat("\n      Reached node: ", node$getName())
    currentVar = node$getVar()
    #cat("\n      Current variable: ", currentVar$getName())

    # The first node include in reachedNodes is the root which is always included
    reachedNodes = c(reachedNodes, node)

    if (assignment$isEvidenced(currentVar) && (!node$isLeaf())) {
      #print(assignment)
      value = assignment$getValue(currentVar)
      #cat("\n      Evidenced value:", value)

      if(node$isCategorical()){
        index = node$getSuccPosition(value)
        #cat("\n      Index position for the successor:", index)
      }

      for(i in 1:currentVar$numChildren()){
        child = currentVar$getChildAt(i)
        #cat("\n Child:", child$getName())
        if(node$isContinuous()){
          index = node$getSuccPosition(child,value)

          #cat("\n      Index position for the successor:", index)
        }
        nextNode = node$succ(child,index)
        reachedNodes = c(reachedNodes,reachedNodesFromNode(nextNode,assignment, c()))
      }

      #x = node$getSuccessorAt(i)
      #cat("\n   Successor: ", x$getName())
      #cat("\n   Index: ", index)
      #nextNode = node$succ(x$getVar(),index)

      #if(x$equals(nextNode)){
      #if(identical(x,nextNode)){
      #cat("\n   Recursive call with node",nextNode$getName())
      #  reachedNodes = c(reachedNodes,reachedNodesFromNode(nextNode,assignment, c()))
      #}
      #else
      #  cat("\nNo more recursive calls")
    }
    return(reachedNodes)
  }
)

##################################################################################################

# Returns the parameter nodes reached by a given assignment bellow THIS parameter node.
#
# Args:
#   assignment: An assigment.
#
# Returns:
#   The reached parameter nodes by the assignment.

ParameterNode$methods (
  reachedNodes = function(assignment) {
    reachedNodesFromNode(.self,assignment, c())
  }
)

##################################################################################################

# ParameterNode$methods (
#   evaluate = function(value) {
#     pos = getSuccPosition(value)
#     return(getDistribution()$getParameters()[pos])
#   }
# )

##################################################################################################

# Gets the number of cases reached by this parameter node.
#
# Args:
#
# Returns:
#   The number of cases reached by this parameter node.

ParameterNode$methods (
  numReachedCases = function() {
    getReachedData()$getNumberOfCases()
  }
)

##################################################################################################

# Gets the number of parameters of its distribution.
#
# Args:
#
# Returns:
#   The number of parameters of this parameter node.

ParameterNode$methods (
  numParameters = function() {
    .self$getDistribution()$numParameters()
  }
)

##################################################################################################

# Gets the position(s)of a parameter in the list of successors.
#
# Args:
#   param: The parameter to be checked.
#
# Returns:
#   The position of a param in the list of successors.

ParameterNode$methods (
  getPositionsSucc = function(succ) {

    vPos = c()
    for (i in 1:.self$numSuccessors()){
      if(identical(getSuccessorAt(i),succ)){
        vPos=c(vPos,i)
      }
    }
    if (length(vPos)==0) stop("\n Parameter does not exist")
    return(vPos)

  }
)

##################################################################################################

# Gets the successor names of this parameter node.
#
# Args:
#
# Returns:
#   A vector with the successor names of this parameter node.

ParameterNode$methods (
  getSuccessorNames = function() {
    if(.self$numSuccessors()==0) return(NULL)
    vNames=c()
    for (i in 1:.self$numSuccessors()){
      vNames = c(vNames,getSuccessorAt(i)$getName())
    }
    return(vNames)
  }
)

##################################################################################################

# Gets the predecessor names of this parameter node.
#
# Args:
#
# Returns:
#   A vector with the predecessor names of this parameter node.

ParameterNode$methods (
  getPredecessorNames = function() {
    if(.self$numPredecessors()==0) return(NULL)
    vNames=c()
    for (i in 1:.self$numPredecessors()){
      vNames = c(vNames,getPredecessorAt(i)$getName())
    }
    return(vNames)
  }
)

##################################################################################################

# Gets the number of child variables of this parameter.
#
# Args:
#
# Returns:
#   The number of children of this parameter.

ParameterNode$methods (
  numChildren = function() {
    getVar()$numChildren()
  }
)

##################################################################################################

# Gets the child variables of this parameter.
#
# Args:
#
# Returns:
#   The children of this parameter.

ParameterNode$methods (
  getChildren = function() {
    getVar()$getChildren()
  }
)

##################################################################################################
######################################## CLASS DEFINITION ########################################
##################################################################################################
CategoricalParameterNode <- setRefClass(
  Class = "CategoricalParameterNode",
  fields = list(
  ),
  contains="ParameterNode"
)

##################################################################################################
###################################### SETTERS AND GETTERS #######################################
##################################################################################################

##################################################################################################
########################################### METHODS ##############################################
##################################################################################################

# Initializes the categorical parameter node with a name, a distribution and a reached data.
#
# Args:
#   name: The name of the parameter node (non mandatory argument).
#   distribution: A categorical potential.
#   reachedData: A Database with the reached data of this parameter.
#
# Returns:
#   The initialized CategoricalParameterNode object.

CategoricalParameterNode$methods (
  initialize = function(name=NULL,distribution=NULL, reachedData=NULL){
    if (!is.null(name)) setName(name)
    if (!is.null(distribution)) setDistribution(distribution)
    if(!is.null(reachedData)) setReachedData(reachedData)
  }
)

##################################################################################################

# Shows relevant information about the parameter node: name, distribution and reached cases.
#
# Args:
#
# Returns:
#   No arguments are returned. Only the information is displayed on the screen.

CategoricalParameterNode$methods (
  show = function() {
    cat("Parameter:", getName(),"\n")
    if (getDistribution()$numParameters()>0)
      print(getDistribution())
    cat("\nReached cases:",getReachedData()$getNumberOfCases(),"\n")
  }
)

##################################################################################################

# Gets the position (in the list of successors) in which the parameter node corresponding
# to a given value (state) is stored.
#
# Args:
#   value: A state of the variable to which this parameter node belongs.
#
# Returns:
#   A integer in the set [1,2,...,numStates] with the position in which the value is stored.

CategoricalParameterNode$methods (
  getSuccPosition = function(value) {
    for (i in 1:.self$getVar()$numStates()){
      if(value==.self$getVar()$getState(i)){
        return(i)
      }
    }
    stop("The value for the variable is not valid.")
  }
)

##################################################################################################

# Gets the probability of a given state.
#
# Args:
#   value: A state of the variable to which this parameter node belongs.
#
# Returns:
#   The probability of a given state.

CategoricalParameterNode$methods (
  evaluate = function(value) {
    pos = getSuccPosition(value)
    return(getDistribution()$getParameters()[pos])
  }
)

##################################################################################################

# Gets the size of this parameter node, i.e., the number of parameters - 1.
#
# Args:
#
# Returns:
#   The size of the parameter.
#

CategoricalParameterNode$methods (
  size = function() {
    numParameters()-1
  }
)

##################################################################################################
######################################## CLASS DEFINITION ########################################
##################################################################################################
ContinuousParameterNode <- setRefClass(
  Class = "ContinuousParameterNode",
  fields = list(
    maxIntervals = "numeric",
    cutPoints="list"
  ),
  contains="ParameterNode"
)

##################################################################################################
###################################### SETTERS AND GETTERS #######################################
##################################################################################################
ContinuousParameterNode$accessors(c("maxIntervals","cutPoints"))

##################################################################################################
########################################### METHODS ##############################################
##################################################################################################

# Initializes the continuous parameter node with a name, an a set of cut points.
#
# Args:
#   name: The name of the parameter node.
#   cutPoints: A vector of cut points of the domain of the variable.
#
# Returns:
#   The initialized ContinuousParameterNode object.

ContinuousParameterNode$methods (
  initialize = function(maxIntervals,name=NULL, cutPoints = NULL){
    setMaxIntervals(maxIntervals)
    if (!is.null(name)) setName(name)
    if (!is.null(cutPoints)) setCutPoints(cutpoints)
    else {
      setCutPoints(list())
    }
  }
)

##################################################################################################

# Computes the number of cutpoints of this parameter node.
#
# Args:
#
# Returns:
#   The number of cutpoints of this parameter node.

ContinuousParameterNode$methods (
  numCutPoints = function(){
    length(getCutPoints())
  }
)

##################################################################################################

# Shows relevant information about the parameter node: name, distribution, reached cases
# and cutpoints
#
# Args:
#
# Returns:
#   No arguments are returned. Only the information is displayed on the screen.

ContinuousParameterNode$methods (
  show = function() {
    cat("Parameter:", getName(),"\n")
    if (getDistribution()$numParameters()>0){
      print(getDistribution())
      cat("\nReached cases:",getReachedData()$getNumberOfCases())

      var = getVar()
      cat("\n    Cutpoints:")
      #cat("\nDATA:\n")
      #print(getReachedData()$getData())
      if(length(cutPoints)>0){
        for (i in 1:var$numChildren())
          cat("\nChild:",var$getChildAt(i)$getName(),":", cutPoints[[i]])

      }
      cat("\n")
    }
  }
)

##################################################################################################

#ARREGLAR Gets the position (in the list of successors) in which the parameter node corresponding
# to a given value is stored.
#
# Args:
#   value: A continuous value of the variable to which this parameter node belongs.
#
# Returns:
#   A integer in the set [1,...,numCutPoints+1] with the position for the value. Note that
#   max and min are not included in the set of cutpoints.

ContinuousParameterNode$methods (
  getSuccPosition = function(child,value) {

    var = getVar()
    if((value>var$getMax()) || (value<var$getMin()))  stop("Value out of range")

    i = var$getChildPosition(child)
    vCutPoints = cutPoints[[i]]

    if (length(vCutPoints)==0) return(1)


    for(i in 2:(length(vCutPoints))){
      if(vCutPoints[i]>value){
        return(i-1)
      }
    }
    # if the value is equal to the last cutpoint, we need this return.
    return(i-1)
  }
)

##################################################################################################

# Evaluate the density function of this parameter node for a given value.
#
# Args:
#   value: A continuous value
#
# Returns:
#   The value of the density for the given argument.

ContinuousParameterNode$methods (
  evaluate = function(value) {
    # Comprobar si value se sale del dominio de este parametro!!!!!!!!!
    getDistribution()$evaluate(value)
  }
)

##################################################################################################

# Adds a vector of cutpoints to the list of cutpoints of this parameter. There are two scenarios:
# 1) The current variable has no children, that is, listCutPoints is empty. In this case, the
#    new cutpoints are computed using the equal frequency method.
# 2) The current variable has at least one child. The new vector of cutpoints is copied from the
#    former one. The list will have as many vectors as children, all of them equals. After building
#    the PDG and the merge operation is carried out, the cutpoints may change for different
#    child variables.
#
# Args:
#
# Returns:
#   A list with vectors of cutpoints, one for each child variable.
#
ContinuousParameterNode$methods (
  addVectorOfCutPointsEF = function() {
    if(length(cutPoints)==0){
      a = getVar()$getMin()
      b = getVar()$getMax()
      data = getReachedData()$getColumnData(getVar())
      vPoints = as.numeric(names(table(quantile(data, probs=seq(0,1,by=1/getMaxIntervals())))))
      if(length(vPoints)<getMaxIntervals()+1) {
        vPoints = seq(a,b,by = (b-a)/getMaxIntervals())
      }
      vPoints[1]=a
      vPoints[getMaxIntervals()+1]=b
      cutPoints[[length(cutPoints)+1]] <<- vPoints
    }
    else{
      cutPoints[[length(cutPoints)+1]] <<- cutPoints[[length(cutPoints)]]
    }
    return(cutPoints)
  }
)

##################################################################################################

# Gets the size of this parameter node, i.e., the number of parameters - 1.
# For continuous parameters, we also add the number of cutpoints in case of having children.
#
# Args:
#
# Returns:
#   The size of the parameter.
#

ContinuousParameterNode$methods (
  size = function() {
    s = numParameters()-1
    if(!isLeaf())
      s = s + sum(sapply(getCutPoints(), function(v){length(v)-2}))
    return(s)
  }
)

#l = list(c(1,2,3),c(1,4,5,6,7,8))
#sum(sapply(l, function(v){length(v)-2}))

##################################################################################################

# Gets the density type of this parameter node.
#
# Args:
#
# Returns:
#   A String indicating the density type of the parameter ("MOP" or "MTE").

ContinuousParameterNode$methods (
  getDensityType = function() {
    getDistribution()$getType()
  }
)
